//====----------------------- --- --   -      -
// MULTIUSER NETWORK MANAGEMENT
// TCP/IP and UDP support.
// 2000-2002, Vander Nunes
//======----------------------------- --- --   -      -


#include "http.h"
#include "netclient.h"



// ------------------------------------------------------------------------------------------------
//
// Http Window Procedure
//
// ------------------------------------------------------------------------------------------------
LRESULT CALLBACK HttpWndProc(HWND hWnd, UINT uMessage, WPARAM wParam, LPARAM lParam)
{

  // get the pointer to the netserver class
  CHttp *pHttp = (CHttp*)GetProp(hWnd, "http");
  if (!pHttp)
    return DefWindowProc(hWnd,uMessage,wParam,lParam);

  // handle the message
  switch (uMessage)
  {

    case WM_TIMER:
    {
      // connection timed out.
      if (pHttp->m_Stage == dsConn || pHttp->m_Stage == dsReq)
        pHttp->Fail();

      break;
    }

  }

  // return default window procedure result
  return DefWindowProc(hWnd,uMessage,wParam,lParam);
}



// ----------------------------------------------------------------------------
//
// Private callback for the socket client - just pass to the class callback
//
// ----------------------------------------------------------------------------
void HttpClientCallback(CNetClient *pClient, void *pData, int iMsg)
{
  CHttp *pHttp = (CHttp*)pData;
  pHttp->Callback(pClient, iMsg);
}



// ------------------------------------------------------------------------------------------------
//
// Constructor
//
// User instantiating the class can provide a callback to receive http events,
// and store a pointer to any data wanted. This data will be repassed to the callback.
//
// ------------------------------------------------------------------------------------------------
CHttp::CHttp(USR_HTTPCALLBACK(userProc), void *pUserData)
{
  m_hFile = NULL;
  m_pClient = NULL;
  m_Stage = dsIdle;
  m_dwBytesDownloaded = 0;
  m_userProc = userProc;
  m_pUserData = pUserData;
  m_bWorking = FALSE;

  m_bUseProxy = FALSE;
  m_szProxyAddr[0] = 0;
  m_iProxyPort = 0;

  m_hWnd = NetMakeWindow("http",1,64,64,WS_OVERLAPPEDWINDOW,HttpWndProc);
  if (!m_hWnd) throw (1);

  SetProp(m_hWnd,"http",this);
}



// ------------------------------------------------------------------------------------------------
//
// Destructor - closes any pending download before instance destruction.
//
// ------------------------------------------------------------------------------------------------
CHttp::~CHttp()
{
  CloseDownload();
  if (m_hWnd) DestroyWindow(m_hWnd);
}



// ------------------------------------------------------------------------------------------------
//
// Start a download.
// If you need multiple downloads, use multiple instances of the class.
//
// ex: pHttp->Download("www.virtware.net", "/downs/techdemo.exe", "c:\\downs\\techdemo.exe");
//
// ------------------------------------------------------------------------------------------------
BOOL CHttp::Download(char *szHost, char *szFile, char *szTarget, int iPort, char *szProxyAddr, int iProxyPort)
{
  // refuses to start a new download if a download is happening.
  if (m_bWorking) return FALSE;

  // safe check names
  if (strlen(szHost)>=MAX_PATH || strlen(szFile)>=MAX_PATH || strlen(szTarget)>=MAX_PATH)
    return FALSE;

  if (szProxyAddr)
  {
    if (strlen(szProxyAddr) > sizeof(m_szProxyAddr))
      return FALSE;
  }

  // check if there is proxy information
  if (szProxyAddr && iProxyPort)
  {
    strcpy(m_szProxyAddr, szProxyAddr);
    m_iProxyPort = iProxyPort;
    m_bUseProxy = TRUE;
  }
  else
    m_bUseProxy = FALSE;

  try
  {
    m_pClient = new CNetClient(stTCP, HttpClientCallback, this, BUFFER_SIZE);

    if (!m_bUseProxy)
      m_pClient->Connect(szHost, iPort);
    else
      m_pClient->Connect(szProxyAddr, iProxyPort);
  }
  catch (...)
  {
    // failed to initialize CNetClient
    // (maybe because already downloading something?)
    return FALSE;
  }

  m_Stage = dsConn;
  strcpy(m_szHost, szHost);
  strcpy(m_szFile, szFile);
  strcpy(m_szTarget, szTarget);
  m_dwBytesDownloaded = 0;

  m_bWorking = TRUE;

  // callback user function
  if (m_userProc)
    m_userProc(this, m_pUserData, m_Stage, m_dwBytesDownloaded);

  // set a timer
  // to timeout the connection+request stages.
  SetTimer(m_hWnd, 1, 60000, NULL);

  return TRUE;
}



// ------------------------------------------------------------------------------------------------
//
// Private http callback - handles socket connection and data receiving
//
// ------------------------------------------------------------------------------------------------
void CHttp::Callback(CNetClient *pClient, int iMsg)
{
  char szTmp[512];

  switch (iMsg)
  {

    case NET_CONNECTED:
    {
      // send GET request to server
      if (!m_bUseProxy)
        sprintf(szTmp, HTTP_GETREQUEST, "", "", m_szFile, m_szHost);
      else
        sprintf(szTmp, HTTP_GETREQUEST, "http://", m_szHost, m_szFile, m_szHost);

      pClient->Send((void*)szTmp,strlen(szTmp)+1);

      // change stage
      m_Stage = dsReq;

      // callback user function
      if (m_userProc)
        m_userProc(this, m_pUserData, m_Stage, m_dwBytesDownloaded);

      break;
    }

    case NET_NEWDATA:
    {

      NETPACKET *pPacket = pClient->Read();

      switch (m_Stage)
      {

        // ----------------------------------------------------------------------------------------
        // Requesting
        // ----------------------------------------------------------------------------------------
        case dsReq:
        {

          // kill the time out timer
          KillTimer(m_hWnd, 1);

          // expecting server response to GET request
          if ( strpos(HTTP_GETOK, (char*)pPacket->pBuffer) >= 0 )
          {
            // open the target file
            m_hFile = fopen(m_szTarget, "wb+");
            if (!m_hFile)
            {
              // failed to open target file
              Fail();
              return;
            }

            // find the end of the answer header
            // (the start of the file can be glued at the end of the header)
            int iFirst = strpos(HTTP_HEADEREND, (char*)pPacket->pBuffer);

            // if found the end, see if there is something to write already
            if (iFirst >= 0)
            {
              // skip the header end (will point at the first byte of the file or end of the block)
              iFirst += strlen(HTTP_HEADEREND);

              // check to see if there is data remaining to write yet
              if (pPacket->iLength > iFirst)
              {
                // there is file data already, write it
                m_dwBytesDownloaded = pPacket->iLength-iFirst;
                BYTE *p = (BYTE*)pPacket->pBuffer;
                fwrite(&p[iFirst], 1, m_dwBytesDownloaded, m_hFile);
              }
            }

            // change stage
            m_Stage = dsRecv;

            // callback user function
            if (m_userProc)
              m_userProc(this, m_pUserData, m_Stage, m_dwBytesDownloaded);
          }
          else
          {
            if (pPacket->iLength < BUFFER_SIZE)
            {
              // failed
              Fail();
            }
          }

          break;
        }
        // ----------------------------------------------------------------------------------------


        // ----------------------------------------------------------------------------------------
        // Downloading
        // ----------------------------------------------------------------------------------------
        case dsRecv:
        {

          do
          {

            if (pPacket->iLength > 0)
            {
              // write to file
              fwrite(pPacket->pBuffer, 1, pPacket->iLength, m_hFile);

              // update bytes downloaded
              m_dwBytesDownloaded += pPacket->iLength;

              // callback user function
              if (m_userProc)
                m_userProc(this, m_pUserData, m_Stage, m_dwBytesDownloaded);
            }

            // get new packet
            pPacket = pClient->Read();

          } while (pPacket->iLength > 0);

          // check if there was an error code
          if (pPacket->iLength == SOCKET_ERROR)
          {
            int nRc = WSAGetLastError();

            // check why we received an error code
            if (nRc != WSAEWOULDBLOCK)
            {
              // finished - ### guess if this happens here, we failed to download?
              // Fail();
              Stop();
            }
          }

          break;

        }
        // ----------------------------------------------------------------------------------------

      }

      break;
    }

    case NET_DISCONNECTION:
    {
      // finished
      Stop();

      break;
    }

  }
}



// ------------------------------------------------------------------------------------------------
//
// Just close any pending download - doesn't change stage nor call user.
//
// ------------------------------------------------------------------------------------------------
void CHttp::CloseDownload()
{
  if (m_hFile)
  {
    fclose(m_hFile);
    m_hFile = NULL;
  }

  if (m_pClient)
  {
    delete m_pClient;
    m_pClient = NULL;
  }

  m_bWorking = FALSE;
}



// ------------------------------------------------------------------------------------------------
//
// Stop a download (with success)
//
// ------------------------------------------------------------------------------------------------
void CHttp::Stop()
{
  CloseDownload();

  m_Stage = dsIdle;

  // callback user function
  if (m_userProc)
    m_userProc(this, m_pUserData, m_Stage, m_dwBytesDownloaded);
}



// ------------------------------------------------------------------------------------------------
//
// Stop a download (with fail)
//
// ------------------------------------------------------------------------------------------------
void CHttp::Fail()
{
  CloseDownload();

  m_Stage = dsFail;

  // callback user function
  if (m_userProc)
    m_userProc(this, m_pUserData, m_Stage, m_dwBytesDownloaded);
}



// ------------------------------------------------------------------------------------------------
//
// Private helper method - tell where inside szString is located szStr
//
// ------------------------------------------------------------------------------------------------
int CHttp::strpos(char *szStr, char *szString)
{
  DWORD p = 0;

  for (DWORD i=0; i<strlen(szString); i++)
  {
    if (szString[i] == szStr[p])
    {
      p++;
      if (p >= strlen(szStr))
        return i-(p-1);
    }
    else p = 0;
  }

  return -1;
}
