//====----------------------- --- --   -      -
// MULTIUSER NETWORK MANAGEMENT
// TCP/IP and UDP support.
// 2000-2002, Vander Nunes
//======----------------------------- --- --   -      -


#include <windows.h>
#include "netserver.h"


DWORD CNetServer::m_dwRefCount = 0;


// ------------------------------------------------------------------------------------------------
//
// Server Window Procedure
//
// ------------------------------------------------------------------------------------------------
LRESULT CALLBACK NetWndProc(HWND hWnd, UINT uMessage, WPARAM wParam, LPARAM lParam)
{
  // get the pointer to the netserver class
  CNetServer *pServer = (CNetServer*)GetProp(hWnd, "netserver");

  // handle the message
  switch (uMessage)
  {
    case NET_SELECT:
    {
      pServer->Select(wParam, lParam);
      break;
    }
  }

  // return default window procedure result
  return DefWindowProc(hWnd,uMessage,wParam,lParam);
}



// ------------------------------------------------------------------------------------------------
//
// Construction
//
// ------------------------------------------------------------------------------------------------
CNetServer::CNetServer(enSockType SockType, USR_SRVRCALLBACK(userProc), void *pUserData, int iListenPort, int iMaxClients, int iBufferSize, int iTimeout)
{
  m_bInitialized = FALSE;
  m_dwRefCount++;

  // try to initialize buffers
  if ( !Init(iMaxClients, iBufferSize) ) throw (0);

  // try to create hidden window
  if ( !(m_hWnd = NetMakeWindow("netserver",m_dwRefCount,64,32,WS_OVERLAPPEDWINDOW,NetWndProc)) ) throw (1);

  // put our address in the window properties
  if ( !SetProp(m_hWnd, "netserver", (HANDLE)this) ) throw (2);

  m_bInitialized = TRUE;

  // try to start listening for connections
  if ( !Listen(SockType, userProc, pUserData, iListenPort, iTimeout) ) throw (3);
}



// ------------------------------------------------------------------------------------------------
//
// Initialization
//
// ------------------------------------------------------------------------------------------------
BOOL CNetServer::Init(int iMaxClients, int iBufferSize)
{
  if (m_dwRefCount == 1)
  {
    WSADATA wsaData;
    WORD wVersionRequested = MAKEWORD(1,1);
    WSAStartup(wVersionRequested, &wsaData);
  }

  // clear listening sockets
  m_UDPInSock = INVALID_SOCKET;
  m_TCPInSock = INVALID_SOCKET;

  // initialize clients array
  m_iMaxClients = iMaxClients;
  m_iNumClients = 0;
  m_pClients = new CONN_INFO[m_iMaxClients];
  if (!m_pClients) return FALSE;

  // initialize buffers
  m_iRecvBufferSize = iBufferSize;
  m_pRecvBuffer = new char[m_iRecvBufferSize];
  if (!m_pRecvBuffer) return FALSE;

  // point netpacket buffer to the receive buffer
  m_Packet.pBuffer = m_pRecvBuffer;

  return TRUE;
}



// ------------------------------------------------------------------------------------------------
//
// Close winsock service
//
// ------------------------------------------------------------------------------------------------
CNetServer::~CNetServer()
{
  m_dwRefCount--;

  // stop listening
  StopListening();

  // remove properties from our window
  RemoveProp(m_hWnd, "netserver");

  // destroy the window
  DestroyWindow(m_hWnd);

  if (!m_dwRefCount)
  {
    // close sockets service
    WSACleanup();
  }

  // free resources
  if (m_pClients) delete m_pClients;
  if (m_pRecvBuffer) delete m_pRecvBuffer;
}



// ------------------------------------------------------------------------------------------------
//
// Start listening for client connections
//
// ------------------------------------------------------------------------------------------------
BOOL CNetServer::Listen(enSockType SockType, USR_SRVRCALLBACK(userProc), void *pUserData, int iListenPort, int iTimeout)
{
  int nRc;
  int c;

  // refuses to listen if not initialized
  if (!m_bInitialized) return FALSE;

  // stop any listening mode
  StopListening();

  // store listening information
  m_SockType = SockType;
  m_userProc = userProc;
  m_pUserData = pUserData;
  m_iListenPort = iListenPort;
  m_iTimeout = iTimeout;

  // clear any remaining client
  for(c=0; c<m_iMaxClients; c++)
    m_pClients[c].used = 0;

  // get server address
  memset((LPSTR)&m_MyAddr, 0, sizeof(m_MyAddr));
  m_MyAddr.sin_family = PF_INET;
  m_MyAddr.sin_port = htons(m_iListenPort);

  // --------------------------------------------------------------------------
  // listen for UDP connections
  // --------------------------------------------------------------------------

  if(m_SockType == stUDP)
  {

    // user wants to listen for UDP.
    if(m_UDPInSock == INVALID_SOCKET)
    {

      // its not already running, so start it
      m_UDPInSock = socket(PF_INET, SOCK_DGRAM, IPPROTO_UDP);

      if(m_UDPInSock != SOCKET_ERROR)
      {
        nRc = bind(m_UDPInSock, (struct sockaddr *)&m_MyAddr, sizeof(m_MyAddr));

        //
        // Use WSAAsyncSelect() to notify ourselves of network
        // events.  Since this is UDP, we'll never see FD_CONNECT,
        // FD_ACCEPT, or FD_CLOSE, but we can still register for
        // them.
        //
        
        nRc = WSAAsyncSelect(m_UDPInSock,
                             m_hWnd,                  // our window procedure will be the recipient
                             NET_SELECT,              // message to send
                             FD_READ  |               // tell when new data's here
                             FD_WRITE |               // tell when there's buffer space
                             FD_CONNECT |             // tell when the below connect is done
                             FD_ACCEPT |
                             FD_CLOSE);               // tell if the other side closes

      }
      else
      {
        // error while initializing UDP
        m_UDPInSock = INVALID_SOCKET;
        return FALSE;
      }

    }

  }

  // --------------------------------------------------------------------------
  // listen for TCP connections
  // --------------------------------------------------------------------------

  if(m_SockType == stTCP)
  {

    // user wants to listen for TCP.
    if(m_TCPInSock == INVALID_SOCKET)
    {

      // it's not already running, so start it
      m_TCPInSock = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP);

      if(m_TCPInSock != SOCKET_ERROR)
      {
        nRc = bind(m_TCPInSock, (struct sockaddr *)&m_MyAddr, sizeof(m_MyAddr));

        nRc = listen(m_TCPInSock, 2);

        //
        // Use WSAAsyncSelect() to notify ourselves of network
        // events.  This is TCP, so we could see all of the
        // notifications.  However, since we are listening,
        // we'll never see FD_CONNECT on this socket.
        // them.
        //
        
        nRc = WSAAsyncSelect(m_TCPInSock,
                             m_hWnd,         // our window procedure will be the recipient
                             NET_SELECT,
                             FD_READ  |
                             FD_WRITE |
                             FD_CONNECT |
                             FD_ACCEPT |
                             FD_CLOSE);

      }
      else
      {
        // error while initializing TCP
        m_TCPInSock = INVALID_SOCKET;
        return FALSE;
      }

    }

  }

  return TRUE;
}



// ------------------------------------------------------------------------------------------------
//
// Stop listening
//
// ------------------------------------------------------------------------------------------------
void CNetServer::StopListening()
{
  // turn off UDP if it's running.
  if(m_UDPInSock != INVALID_SOCKET)
  {
    closesocket(m_UDPInSock);
    m_UDPInSock = INVALID_SOCKET;
  }

  // turn off TCP if it's running.
  if(m_TCPInSock != INVALID_SOCKET)
  {
    closesocket(m_TCPInSock);
    m_TCPInSock = INVALID_SOCKET;
  }
}



// ------------------------------------------------------------------------------------------------
//
// Process incoming connections, that can be either TCP or UDP.
// If TCP, accepts connection and increments NumClients.
// If UDP, simply receive the message and close the socket.
//
// ------------------------------------------------------------------------------------------------
BOOL CNetServer::Select(int wParam, int lParam)
{
  int nRc, i;
  unsigned long data = 0; 
  unsigned long four_byte;
  struct sockaddr_in rem_addr;
  int namelen = sizeof(rem_addr), ilen = sizeof(unsigned long);
  int clientidx;

  // get the involved socket
  SOCKET sock = (SOCKET)wParam;

  switch(WSAGETSELECTEVENT(lParam))
  {

    case FD_CONNECT:
    {
      //
      // only TCP connections pass here
      //
      break;
    }

    case FD_ACCEPT:
    {
      //
      // (only TCP connections pass here)
      // a new TCP connection has come in.  Accept it, set
      // up the socket info
      //

      // accept the incoming connection
      sock = accept(sock, NULL, NULL);

      // we must get the address of the peer
      nRc = getpeername(sock, (struct sockaddr *)&rem_addr, (LPINT) &namelen);

      // if don't, refuse connection
      if(nRc == SOCKET_ERROR)
      {
        closesocket(sock);
        // tell user about the refused connection
        if (m_userProc) m_userProc(this, m_pUserData, NET_REFCONNECTION, -1);
        return FALSE;
      }

      //?
      //data = rem_addr.sin_addr.s_addr;

      //
      // if this is a new connection, we won't have a local handle
      // yet.  Just pass it on to the new connection routine
      //
      if ((clientidx = FindClientIdx(sock)) == -1)
      {
        // the method will tell user about the new connection
        NewConnection(sock);
      }
      else
      {
        // we don't need to do anything, just
        // allow the socket to get it's FD_READ
        // which will be next
      }
      break;
    }

    case FD_READ:
    {
      //
      // check if really there is need for reading
      //
      nRc = ioctlsocket(sock, FIONREAD, (u_long *)&data);
      if(nRc == SOCKET_ERROR)
        return FALSE;
      if(data <= 0 )
        return FALSE;

      // TCP doesn't fill in the from address on
      // recvfrom().  Find out if this is UDP or
      // tcp, and use an appropriate mechanism to
      // find out who the other end is.
      
      i = sizeof(int);
      getsockopt(sock, SOL_SOCKET, SO_TYPE, (char *)&nRc, (LPINT)&i);

      if(nRc == SOCK_DGRAM)
      {
        //
        // UDP datagram
        //
        nRc = recvfrom(sock, (LPSTR)&four_byte, (int)ilen, MSG_PEEK, (struct sockaddr *)&rem_addr, (LPINT)&namelen);
        if(nRc == SOCKET_ERROR)
        {
          nRc = WSAGetLastError();
          if(nRc != WSAEMSGSIZE)
            return FALSE;
        }
      }
      else
      {
        //
        // TCP packet
        //
        nRc = getpeername(sock, (struct sockaddr *)&rem_addr, (LPINT)&namelen);
      }
      if(nRc == SOCKET_ERROR)
        return FALSE;
      
      data = rem_addr.sin_addr.s_addr;

      //
      // if it's a new connection, we won't have a local handle
      // yet.  Just pass it on to the new connection routine
      //
      if ((clientidx = FindClientIdx(sock)) == -1)
      {
        // it's a new connection.
        // connect and tell user about the new connection
        NewConnection(sock);
      }
      else
      {
        // it isn't a new connection.
        // tell client there is data waiting
        if (m_userProc) m_userProc(this, m_pUserData, NET_NEWDATA, clientidx);
      }

      break;
    }

    case FD_CLOSE:
    {
      //
      // Only TCP pass here.
      //
      // the other side closed the connection.  close the socket,
      // to clean up internal resources.
      //

      // receive the disconnection message
      recvfrom(sock, (LPSTR)&four_byte, (int)ilen, MSG_PEEK, (struct sockaddr *)&rem_addr, (LPINT)&namelen);

      // find the client connection info
      clientidx = FindClientIdx(sock);
      if (clientidx != -1)
      {
        // free client slot.
        m_pClients[clientidx].used = 0;
        m_pClients[clientidx].sock = INVALID_SOCKET;
      }

      // close socket and update client counting
      closesocket(sock);
      m_iNumClients--;

      // tell user about the disconnection
      if (m_userProc) m_userProc(this, m_pUserData, NET_DISCONNECTION, clientidx);

      break;
    }

  }   // end: switch(WSAGETSELECTEVENT(lParam))

  return TRUE;
}



// ------------------------------------------------------------------------------------------------
//
// There is data waiting to be read.
// Receive that.
// Both TCP/IP and UDP pass here, after been 'selected'.
//
// ------------------------------------------------------------------------------------------------
NETPACKET* CNetServer::ReadMessageFromClient(int iClientIdx)
{
  int nRc;
  struct sockaddr rem_addr;
  int ilen = sizeof(struct sockaddr_in);
  CONN_INFO *pClient;

  // safe checking
  if (iClientIdx < 0 || iClientIdx >= m_iMaxClients) return NULL;

  // get the client
  pClient = &m_pClients[iClientIdx];

  // more safe checking
  if (!pClient->used) return NULL;

  nRc = recvfrom( pClient->sock,
                  m_pRecvBuffer,
                  m_iRecvBufferSize,
                  0,
                  (struct sockaddr *)&rem_addr,
                  (LPINT)&ilen
                );

  m_Packet.iLength = nRc;

  return &m_Packet;

  //### NOTE FOR UDP MESSAGES:
  // i think the best thing to do at this stage is:
  // if the first bytes of the message are not a ClientIdx number,
  // allocate a new Client entry for him and
  // answer to the client with his ClientIdx number.
  // the client them should send all his forthcoming messages
  // with the first bytes containing his ClientIdx id.
  // if his message contains the ClientIdx, we know that client
  // and can process his message.
}



// ------------------------------------------------------------------------------------------------
//
// Register an incoming call as a new connection.
// ### Currently it's being called only from TCP calls by selector,
// but should be called for both.
// to register UDP calls, i need to implement some timing control to
// figure out when the client is not communicating anymore.
//
// ------------------------------------------------------------------------------------------------
BOOL CNetServer::NewConnection(SOCKET ServerSocket)
{
  int clientidx;
  int nRc;
  CONN_INFO *conn_info;
  int sz = sizeof(int);
  int type;
  unsigned long llen;
  int ilen;

  // refuse only if server is full.
  if (m_iNumClients == m_iMaxClients)
  {
    closesocket(ServerSocket);
    // tell user about the refused connection
    if (m_userProc) m_userProc(this, m_pUserData, NET_REFCONNECTION, -1);
    return FALSE;
  }

  // seek for an empty slot to accomodate this client.
  for(clientidx=0; clientidx<m_iMaxClients; clientidx++)
    if (!m_pClients[clientidx].used)
      break;

  // should never i==m_iMaxClients!
  if (clientidx==m_iMaxClients) return FALSE;

  // use this slot.
  conn_info = &m_pClients[clientidx];

  // must determine if we are dealing with a TCP (stream)
  // or UDP (dgram) incoming connection.  We could have just
  // passed it in as a flag, but this works as well.
  nRc = getsockopt(ServerSocket, SOL_SOCKET, SO_TYPE,(LPSTR)(LPINT)&type, (LPINT) &sz);

  if(nRc == SOCKET_ERROR)
  {
    nRc = WSAGetLastError();
    return FALSE;
  }

  //
  // initialize the connection structure
  //

  conn_info->sock = INVALID_SOCKET;

  if(type == SOCK_STREAM)
  {

    //
    // it's a TCP socket.  fill the conn_info structure.
    //
    conn_info->protocol = IPPROTO_TCP;
    ilen = sizeof(struct sockaddr_in);
    nRc = getpeername(ServerSocket, (struct sockaddr *)&(conn_info->rem_addr),(LPINT)&ilen);
    if(nRc == SOCKET_ERROR)
      return FALSE;

    // mark this slot as used.
    m_iNumClients ++;
    conn_info->used = 1;
  }
  else if (type == SOCK_DGRAM)
  {

    //
    // it's a UDP socket. We'll need to do a recvfrom() to
    // get the name of the other side, but leave the data there.
    // that will cause another FD_READ to be sent, and then
    // the message will be displayed in the normal manner
    //

    conn_info->protocol = IPPROTO_UDP;

    // first, get the size of the incoming message
    nRc = ioctlsocket(ServerSocket, FIONREAD, (u_long *)&llen);
    if(nRc == SOCKET_ERROR)
      return FALSE;

    ilen = sizeof(struct sockaddr_in);

    //
    // using MSG_PEEK causes the data to be read and left
    // in the queue, so another FD_READ will be issued.
    // by that time, everything is set up, so the right
    // thing will happen.
    //
    nRc = recvfrom(ServerSocket, m_pRecvBuffer, (int)ilen, MSG_PEEK,
              (struct sockaddr *)&(conn_info->rem_addr),
              (LPINT)&ilen);

    if(nRc == SOCKET_ERROR)
    {
      nRc = WSAGetLastError();
      return FALSE;
    }

    // RecvBuf contains the initial data sent from client,
    // using UDP datagram.

  }

  //
  // Associate the conn_info with it
  //
  conn_info->sock = ServerSocket;
  conn_info->lasttick = GetTickCount();

  // tell user about the new connection
  if (m_userProc) m_userProc(this, m_pUserData, NET_NEWCONNECTION, clientidx);

  return TRUE;
}



// ------------------------------------------------------------------------------------------------
//
// Seek for a client, given it's socket.
//
// ------------------------------------------------------------------------------------------------
int CNetServer::FindClientIdx(SOCKET Socket)
{
  int c;

  // search by socket
  for (c=0; c<m_iMaxClients; c++)
    if (Socket == m_pClients[c].sock)
      return c;

  return -1;
}



// ------------------------------------------------------------------------------------------------
//
// Send a message to a client.
//
// ------------------------------------------------------------------------------------------------
void CNetServer::SendMessageToClient(int ClientIdx, void *Msg, int MsgLen)
{
  CONN_INFO *Client = &m_pClients[ClientIdx];

  if (Client->protocol == IPPROTO_TCP)
  {
    //
    // use TCP/IP send()
    //
    send(Client->sock, (const char *)Msg, MsgLen, 0);
  }
  else
  {
    //
    // use UDP sendto()
    //
    sendto(Client->sock, (const char *)Msg, MsgLen, 0, (struct sockaddr *)&Client->rem_addr, sizeof(Client->rem_addr));
  }
}



// ------------------------------------------------------------------------------------------------
//
// Broadcast a message to all clients
//
// ------------------------------------------------------------------------------------------------
void CNetServer::BroadcastMessage(void *Msg, int MsgLen)
{
  int c;

  for (c=0; c<m_iMaxClients; c++)
  {
    if (m_pClients[c].used)
      SendMessageToClient(c,Msg,MsgLen);
  }
}



// ------------------------------------------------------------------------------------------------
//
// Return a pointer to a client connection info
//
// ------------------------------------------------------------------------------------------------
CONN_INFO*  CNetServer::Client(int iClientIdx)
{
  if (iClientIdx < 0 || iClientIdx >= m_iMaxClients) return NULL;
  return (&m_pClients[iClientIdx]);
}
